var Fd = {
  IMAGE_BLANK: FdFile.getAsDataURL("chrome://fastdial/skin/blank.gif"),
  layout: { top: 20, left: 20, spacing: 5, border: 3,
            titlePadding: 4, minWidth: 100, minHeight: 75 },
  menu: null,
  addMenu: null,
  WIDTH: FdPrefs.getIntPref("width"),
  HEIGHT: FdPrefs.getIntPref("height"),
  currentId: 0,
  pageInfos: new Array(), // cached pageInfos
  prefsWindow: null,
  bundle: null,
  RENDER_TIMEOUT: 500,
  forceDrawPage: false,
  pageHtml: null,

  initialize: function() {
    Fd.initShortcutKeys();

    Fd.menu = $("fastdial-menu");
    Fd.addMenu = $("fastdial-add-menu");
    $("contentAreaContextMenu").addEventListener("popupshowing",
                                                 Fd.onContextShowing, false);
    Fd.bundle = $("fd-bundle");

    gBrowser.addEventListener("load", Fd.onPageLoad, true);

    for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
      Fd.pageInfos[i] = FdPrefs.getPageInfo(i);
      if (Fd.pageInfos[i].refresh > 0) Fd.refreshThumbnail(i);
    }

    // Tabs' "busy" attribute may hasn't been set yet, let some time to pass
    setTimeout(Fd.drawPagesOnStartup, 0);
  },

  drawPagesOnStartup: function() {
    FdUtils.forEachBlankPage(function(doc) {
      var tab = FdUtils.getDocumentTab(doc);
      if (!tab.hasAttribute("busy") && FdUtils.isBlank(doc)) {
        Fd.drawPage(doc);
      }
    });
  },

  initShortcutKeys: function() {
    var modifiers = FdPrefs.getCharPref("shortcutKeys");
    for(var i = 1; i <= 9; i++) {
      var key = $("fastdial-key-" + i);
      key.setAttribute("modifiers", modifiers);
    }
  },

  open: function(event) {
    if (event.button == 0) {
      Fd.forceDrawPage = true;
      FdUtils.openUILink("about:blank");
    }
    else if (event.button == 1) {
      Fd.forceDrawPage = true;
      FdUtils.openUILink("about:blank", "tab");
    }
  },

  onPageLoad: function(event) {
    var doc = event.originalTarget;

    if (FdUtils.getDocumentTab(doc)) { // if top-level document
      if (FdUtils.isBlank(doc) &&
          (Fd.forceDrawPage ||
           gBrowser.browsers.length == 1 ||
           FdPrefs.getBoolPref("showInNewTabs")))
      {
        Fd.forceDrawPage = false;
        Fd.drawPage(doc);
      }
      else {
        // Refresh thumbnail for this page if there's refreshOnVisit specified
        for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
          var pageInfo = Fd.pageInfos[i];
          if (pageInfo.refreshOnVisit &&
              doc.location.href == pageInfo.url)
          {
            Fd.saveDoc(i, doc);
          }
        }
      }
    }
  },

  drawPage: function(doc) {
    if (!Fd.pageHtml) {
      Fd.pageHtml = Fd.createPageHtml();
    }
    doc.body.innerHTML = Fd.pageHtml;
    Fd.resizeDoc(doc);

    doc.addEventListener("click", FdPage.onClick, false);
    doc.addEventListener("mousemove", FdPage.onMouseMove, false);
    doc.addEventListener("mousedown", FdPage.onMouseDown, false);
    doc.addEventListener("mouseup", FdPage.onMouseUp, false);
  },

  invalidatePageHtml: function() {
    Fd.pageHtml = null;
  },

  createPageHtml: function() {
    Fd.WIDTH = FdPrefs.getIntPref("width");
    Fd.HEIGHT = FdPrefs.getIntPref("height");

    var hide = " style=\"display: none;\"";

    var html = "";
    for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
      if (!Fd.pageInfos[i]) Fd.pageInfos[i] = FdPrefs.getPageInfo(i);
      var pageInfo = Fd.pageInfos[i];
      var title = pageInfo.title || "&nbsp;";

      if (i % Fd.WIDTH == 1) html += "<tr>";
      html += "<td><div class=\"container\" id=\"container" + i + "\">" +
              "<div class=\"buttons\" id=\"buttons" + i + "\">" +
              "<div class=\"button clear\" title=\"" +
              Fd.bundle.getString("clear") + "\"></div>" +
              "<div class=\"button refresh\" title=\"" +
              Fd.bundle.getString("refresh") + "\"></div>" +
              "<div class=\"button setURL\" title=\"" +
              Fd.bundle.getString("setURL") + "\"></div>" +
              "</div>" +
              "<div class=\"loading\" id=\"loading" + i + "\"" +
              (pageInfo.loading ? "" : hide) + "></div>" +
              "<a href=\"" + pageInfo.url + "\" id=\"link" + i + "\"" +
              (pageInfo.loading ? hide : "") + ">" +
              "<img id=\"" + i + "\" src=\"" + Fd.getImageURL(i) + "\"></a>" + 
              "<div class=\"title\" id=\"title" + i + "\">" + title + "</div>" +
              "</div></td>";
      if (i % Fd.WIDTH == 0) html += "</tr>";
    }
    return Fd.getTemplate().replace(/%content%/, html);
  },

  openLink: function(event) {
    var id = FdPanel.getId(event.target);
    var pageInfo = Fd.pageInfos[id];
    FdUtils.openUILink(pageInfo.url, pageInfo.openIn || event);
  },

  exchangePanels: function(sourceId, targetId) {
    // Exchange pageInfos
    var targetPageInfo = Fd.pageInfos[sourceId];
    var sourcePageInfo = Fd.pageInfos[targetId];
    targetPageInfo.id = targetId;
    sourcePageInfo.id = sourceId;
    Fd.pageInfos[sourceId] = sourcePageInfo;
    Fd.pageInfos[targetId] = targetPageInfo;
    sourcePageInfo.url ? FdPrefs.savePageInfo(sourcePageInfo)
                       : FdPrefs.clearPageInfo(sourcePageInfo.id);
    targetPageInfo.url ? FdPrefs.savePageInfo(targetPageInfo)
                       : FdPrefs.clearPageInfo(targetPageInfo.id);

    // Exchange files
    FdFile.exchangeImages(sourceId, targetId);

    // Update images
    Fd.updateImage(sourceId);
    Fd.updateImage(targetId);
  },

  getTemplate: function() {
    return FdFile.readURL("chrome://fastdial/content/template.html");
  },

  onContextShowing: function(event) {
    if (event.target.id == "contentAreaContextMenu") {
      var doc = gBrowser.selectedBrowser.contentDocument;
      Fd.menu.hidden = !FdUtils.isBlank(doc) ||
                       !gContextMenu.target ||
                       !(Fd.currentId = FdPanel.getId(gContextMenu.target));

      Fd.addMenu.hidden = FdUtils.isBlank(doc) ||
                          !FdPrefs.getBoolPref("showAddMenu");
    }
    else if (event.target.id == "fastdial-popup") {
      var url = Fd.pageInfos[Fd.currentId].url;
      $("fastdial-popup-refresh").hidden = !url;
      $("fastdial-popup-clear").hidden = !url;
    }
    else if (event.target.id == "fastdial-add-popup") {
      FdUtils.clearElement(event.target);
      for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
        var menuitem = document.createElement("menuitem");
        menuitem.setAttribute("label", i + "." + Fd.pageInfos[i].title);
        menuitem.setAttribute("class", "menuitem-iconic");
        menuitem.setAttribute("image", FdFile.getFaviconURL(Fd.pageInfos[i].url));
        menuitem.setAttribute("oncommand", "Fd.addCurrentPage(" + i + ");");
        event.target.appendChild(menuitem);
      }
    }
  },

  pageInfo: null,

  setURL: function(id) {
    var oldPageInfo = Fd.pageInfos[id];
    Fd.pageInfo = FdUtils.cloneObject(Fd.pageInfos[id]);
    Fd.showSetURL();

    if (Fd.pageInfo) {
      if (Fd.pageInfo.url) {

        Fd.pageInfos[id] = Fd.pageInfo;
        FdPrefs.savePageInfo(Fd.pageInfo);

        if (Fd.pageInfo.url != oldPageInfo.url || !Fd.pageInfo.title ||
            Fd.pageInfo.customImage != oldPageInfo.customImage ||
            Fd.pageInfo.customBackground != oldPageInfo.customBackground)
        {
          Fd.refreshThumbnail(id);
        }
        else {
          if (Fd.pageInfo.title != oldPageInfo.title) {
            Fd.updateImage(id);
          }
          if (Fd.pageInfo.refresh != oldPageInfo.refresh) {
            Fd.updateRefreshTimeout(id);
          }
        }
      } else Fd.clearURL(id, true);
    }
  },

  showSetURL: function() {
    openDialog("chrome://fastdial/content/seturl.xul",
               "fastdial-seturl-window",
               "chrome,centerscreen=yes,modal=yes,resizable=yes");
  },

  refreshThumbnail: function(id) {
    var pageInfo = Fd.pageInfos[id];
    if (!pageInfo.url) return;

    pageInfo.loading = true;
    Fd.updateImage(id);

    if (pageInfo.customImage && pageInfo.title) {
      Fd.saveCustomImage(pageInfo);
    }
    else Fd.saveImage(pageInfo);
  },

  createFrame: function(url) {
    var doc = gBrowser.selectedBrowser.contentDocument;
    // Local URLs cannot be accessed from "about:blank" page, so use hidden window
    if (FdFile.isLocalURL(url) || !FdUtils.isBlank(doc)) {
      doc = FdUtils.getHiddenWindow().document;
    }
    var iframe = doc.createElement("iframe");
    iframe.style.position = "absolute";
    iframe.style.left = -2000; // Position the frame outside the screen
    iframe.style.top = -2000;
    iframe.width = 1024;
    iframe.height = 768;
    doc.body.appendChild(iframe);
    return iframe;
  },

  deleteFrame: function(iframe) {
    iframe.parentNode.removeChild(iframe);
  },

  loadInFrame: function(url, onLoad) {
    function onFrameLoad() {
      clearTimeout(loadTimeout);
      onLoad(iframe);
    }
    var iframe = Fd.createFrame(url);
    iframe.addEventListener("load", onFrameLoad, true);
    var loadTimeout = setTimeout(onFrameLoad,
                                 FdPrefs.getIntPref("maxLoadTime") * 1000);
    iframe.setAttribute("src", url);
  },

  saveImage: function(pageInfo) {
    Fd.loadInFrame(pageInfo.url, function(iframe) {
        if (!pageInfo.title) {
          pageInfo.title = iframe.contentDocument.title;
          FdPrefs.savePageInfo(pageInfo);
        }
        if (pageInfo.customImage) {
          Fd.saveCustomImage(pageInfo);
        }
        else Fd.saveThumbnail(pageInfo.id, iframe);
      });
  },

  saveCustomImage: function(pageInfo) {
    // "load" event doesn't fire for cached images, so work around cache
    var url = FdFile.getFileURL(pageInfo.customImage) + "?" + new Date().getTime();

    Fd.loadInFrame(url, function(iframe) {
        var doc = iframe.contentDocument;
        var img = doc.body.firstChild;
        if (FdThumbnail.getHeight(img.width) / img.height > 1) {
          iframe.width = img.width;
          iframe.height = FdThumbnail.getHeight(img.width);
        }
        else {
          iframe.width = FdThumbnail.getWidth(img.height);
          iframe.height = img.height;
        }
        doc.body.style.background = pageInfo.customBackground || "white";
        doc.body.style.margin = 0;
        doc.body.style.width = iframe.width;
        doc.body.style.height = iframe.height;
        doc.body.style.display = "table-cell";
        doc.body.style.textAlign = "center";
        doc.body.style.verticalAlign = "middle";

        Fd.saveThumbnail(pageInfo.id, iframe);
      });
  },

  saveThumbnail: function(id, iframe) {
    Fd.pageInfos[id].loading = false;
    Fd.pageInfos[id].imageURL = null;

    setTimeout(function() {
        var thumbWidth = FdPrefs.getIntPref("thumbWidth");
        var image = FdThumbnail.createThumbnail(iframe.contentWindow, thumbWidth);

        Fd.deleteFrame(iframe);

        FdFile.saveImage(id, image);
        Fd.updateImage(id);
        Fd.updateRefreshTimeout(id);
      },
      Fd.RENDER_TIMEOUT);
  },

  saveDoc: function(id, doc) {
    var iframe = Fd.createFrame();
    iframe.addEventListener("load", function() {
        Fd.saveThumbnail(id, iframe);
      },
      true);

    var nsiUrl = Components.classes["@mozilla.org/network/standard-url;1"]
               .createInstance(Components.interfaces.nsIURL);
    nsiUrl.spec = doc.location.href;
    nsiUrl.fileName = "";

    iframe.src = "data:text/html,<base href=\"" + nsiUrl.spec + "\">" +
                 "<META http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">" +
                 doc.documentElement.innerHTML;
  },

  updateImage: function(id) {
    FdUtils.forEachBlankPage(function(doc) {
      var image = doc.getElementById(id);
      if (image) {
        var pageInfo = Fd.pageInfos[id];
        FdFile.removeFromCache(pageInfo.imageURL);
        image.src = Fd.getImageURL(id);

        var panel = new FdPanel(image);
        var link = panel.get("link");
        link.href = pageInfo.url;
        link.style.display = pageInfo.loading ? "none" : "block";

        panel.get("loading").style.display = pageInfo.loading ? "block" : "none";
        panel.get("title").innerHTML = pageInfo.title || "&nbsp;";
      }
    });
    Fd.invalidatePageHtml();
  },

  clearURL: function(id, dontAskConfirmation) {
    var pageInfo = Fd.pageInfos[id];
    if (!pageInfo.url) return;

    var title = pageInfo.title || pageInfo.url;
    var message = Fd.bundle.getFormattedString("clearConfirmation", [title]);
    if (dontAskConfirmation || FdUtils.confirm(message)) {
      clearTimeout(pageInfo.timeout);
      FdPrefs.clearPageInfo(id);
      Fd.pageInfos[id] = FdPrefs.getPageInfo(id); // update pageInfo cache
      FdFile.deleteImage(id);
      Fd.updateImage(id);
    }
  },

  getImageSize: function() {
    var width, height;

    if (FdPrefs.getBoolPref("fixedSize")) {
      width = FdPrefs.getIntPref("thumbWidth");
      height = FdThumbnail.getHeight(width);
    } else {
      var wnd = gBrowser.selectedBrowser.contentWindow;
      width = (wnd.innerWidth - (Fd.layout.left*2 - Fd.layout.spacing)) /
              Fd.WIDTH - (Fd.layout.border*2 + Fd.layout.spacing);
      height = (wnd.innerHeight - (Fd.layout.top*2 - Fd.layout.spacing)) /
               Fd.HEIGHT - (Fd.layout.border*2 + Fd.layout.spacing +
               Fd.layout.titlePadding + FdPrefs.getFont().size);
    }
    width = Math.max(Math.floor(width), Fd.layout.minWidth);
    height = Math.max(Math.floor(height), Fd.layout.minHeight);

    if (FdThumbnail.getHeight(width) / height > 1) {
      width = FdThumbnail.getWidth(height);
    }
    else height = FdThumbnail.getHeight(width);

    return { width: width, height: height };
  },

  resize: function() {
    FdUtils.forEachBlankPage(Fd.resizeDoc);
  },

  resizeDoc: function(doc) {
    var imageSize = Fd.getImageSize();

    if (doc.styleSheets.length > 0) {
      var rules = doc.styleSheets[0].cssRules;
      for(var j in rules) {
        switch(rules[j].selectorText) {
          case "body":
              var font = FdPrefs.getFont();
              rules[j].style.fontFamily = font.family;
              rules[j].style.fontSize = font.size;
              break;
          case ".loading":
          case "img":
              rules[j].style.height = imageSize.height;
              rules[j].style.width = imageSize.width;
              break;
          case ".title":
              rules[j].style.width = imageSize.width;
              break;
        }
      }
    }
  },

  openPreferences: function() {
    if (!Fd.prefsWindow || Fd.prefsWindow.closed) {
      Fd.prefsWindow = window.openDialog("chrome://fastdial/content/preferences.xul",
                                         "fastdial-preferences-window",
                                         "chrome,centerscreen,resizable=yes");
    } else Fd.prefsWindow.focus();
  },

  getImageURL: function(id) {
    var pageInfo = Fd.pageInfos[id];

    if (pageInfo.url && pageInfo.imageURL == null) {
      pageInfo.imageURL = FdFile.getImageURL(id);
    }
    var url = pageInfo.imageURL || Fd.IMAGE_BLANK;
    return url;
  },

  openPage: function(id) {
    var pageInfo = FdPrefs.getPageInfo(id);
    FdUtils.openUILink(pageInfo.url);
  },

  updateRefreshTimeout: function(id) {
    var pageInfo = Fd.pageInfos[id];
    clearTimeout(pageInfo.timeout);
    if (pageInfo.refresh > 0) {
      pageInfo.timeout = setTimeout(function() { Fd.refreshThumbnail(pageInfo.id); },
                                    pageInfo.refresh * 60 * 1000);
    }
  },

  addCurrentPage: function(id) {
    var doc = gBrowser.selectedBrowser.contentDocument;
    Fd.pageInfo = FdPrefs.createPageInfo(id, doc.location.href, doc.title);
    Fd.showSetURL();

    if (Fd.pageInfo) {
      if (Fd.pageInfo.url) {
        FdPrefs.savePageInfo(Fd.pageInfos[id] = Fd.pageInfo); // assignment
        if (Fd.pageInfo.url != doc.location.href) Fd.refreshThumbnail(id);
        else Fd.saveDoc(id, gBrowser.selectedBrowser.contentDocument);
      }
      else Fd.clearURL(id, true);
    }
  }
}
