var FdFile = {
  getImagesDirectory: function() {
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
               .getService(Components.interfaces.nsIProperties)
               .get("ProfD", Components.interfaces.nsIFile);
    file.append("fastdial");
    if (!file.exists()) file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0777);
    return file;
  },

  getImageName: function(id) {
    return id + ".png";
  }, 

  getImageFile: function(id) {
    var file = FdFile.getImagesDirectory();
    file.append(FdFile.getImageName(id));
    return file;
  },

  deleteImage: function(id) {
    var file = FdFile.getImagesDirectory();
    file.append(FdFile.getImageName(id));
    try {
      file.remove(false);
    } catch(e) {}
  },

  // "chrome://" images in FF 2.x expire from cache after 1 minute,
  // so use "data:" urls to avoid reloading of images

  getImageURL: function(id) {
    if (FdUtils.isFirefox3())  {
      return "chrome://fastdial-images/content/" + FdFile.getImageName(id);
    }
    try {
      var file = FdFile.getImageFile(id);
      var url = FdFile.getFileURL(file);
      return FdFile.getAsDataURL(url);
    }
    catch(e) { return ""; }
  },

  readURL: function(url) {
    var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                    .getService(Components.interfaces.nsIIOService);
    var bstream = Components.classes["@mozilla.org/binaryinputstream;1"]
                  .createInstance(Components.interfaces.nsIBinaryInputStream);
    var channel = ioService.newChannel(url, null, null);
    var input = channel.open();
    bstream.setInputStream(input);
    var contents = bstream.readBytes(bstream.available());
    bstream.close();
    input.close();
    return contents;
  },

  saveImage: function(id, data) {
    var file = FdFile.getImageFile(id);
    FdFile.writeFile(file, data);
  },

  writeFile: function(file, data) {
    var out = Components.classes["@mozilla.org/network/file-output-stream;1"]
              .createInstance(Components.interfaces.nsIFileOutputStream);
    out.init(file, 0x04 | 0x08 | 0x20, 0666, 0); // read & write, create, truncate
    out.write(data, data.length);
    out.close();
  },

  getAsDataURL: function(url) {
    var data = FdFile.readURL(url);
    return "data:image/png;base64," + btoa(data);
  },

  exchangeImages: function(sourceId, targetId) {
    var sourceFile = FdFile.getImageFile(sourceId);
    var sourceExists = sourceFile.exists();
    var targetFile = FdFile.getImageFile(targetId);
    var targetExists = targetFile.exists();
    if (sourceExists && targetExists) {
      sourceFile.moveTo(null, "tmp" + new Date().getTime());
    }
    if (targetExists) targetFile.moveTo(null, FdFile.getImageName(sourceId));
    if (sourceExists) sourceFile.moveTo(null, FdFile.getImageName(targetId));
  },

  getFileURL: function(file) {
    var nsiFile;
    if (file instanceof Components.interfaces.nsIFile) nsiFile = file;
    else {
      nsiFile = Components.classes["@mozilla.org/file/local;1"]
                .createInstance(Components.interfaces.nsILocalFile);
      nsiFile.initWithPath(file);
    }
    var ios = Components.classes["@mozilla.org/network/io-service;1"]
              .getService(Components.interfaces.nsIIOService);
    var fileHandler = ios.getProtocolHandler("file")
                     .QueryInterface(Components.interfaces.nsIFileProtocolHandler);
    return fileHandler.getURLSpecFromFile(nsiFile);
  },

  getFaviconURL: function(url) {
    if (!url) return null;

    var nsiUrl = Components.classes["@mozilla.org/network/standard-url;1"]
                 .createInstance(Components.interfaces.nsIURL);
    nsiUrl.spec = url;
    return nsiUrl.prePath + "/favicon.ico";
  },

  isLocalURL: function(url) {
    return url && url.match(/^(file:|chrome:)/);
  },

  removeFromCache: function(url) {
    var cacheSession = Components.classes["@mozilla.org/network/cache-service;1"]
                       .getService(Components.interfaces.nsICacheService)
                       .createSession("image-chrome", 0, false);
    try {
      var entry = cacheSession.openCacheEntry(url,
                    Components.interfaces.nsICache.ACCESS_READ, false);
      entry.doom();
    }
    catch(e) {}
  },
}
